/*============================================================ 
estimation of WSTAR model for
dGDP = f(dCLI)

this code assumes that we have a quarterly series
for the dependent variable, Y, and monthly series
for explanatory variable, X.

example used here is: 
UK
quarterly GDP data
monthly yield curve data

this models dgdp (quarterly obs) as a linear function of lagged GDP and
  lagged yield curve slope (monthly) and nonlinear transition as
  function of 
  +++++++++++++++++
  ysl up to 12 lags
  +++++++++++++++++

  as a result it saves the following information in 
  Gauss matrices
  - RSS, beta, weights, f, confidence intervals

(c) Ralf Becker, 5 Oct 2009

============================================================*/   

library cmlmt,pgraph;
#include cmlmt.sdf

graphset;
gausset;

cls;
format /mb1 /ros 16,8;
format /mat /sa 8,6;

            /* set a random seed - as needed in starting parameter search */

rndseed(1234576);

            /* import data */
            /* UK GDP, quartely, Q1 1955 to Q2 2006 */
            /* 206 observations */
load gdp[] = c:\YOURPATH\ukgdp.txt;
            /* calculate log differences */
dgdp   = 100 * (ln(trimr(gdp,1,0))-ln(trimr(gdp,0,1)));     /* first obs 60q1 */


            /* UK interest data, monthly, Jan 1960 to Oct 2006 */
            /* Col1: 3mths Tbill rate, Col2: 10yr Govt securities */
            /* 562 observations */
load inter[562,2] = c:\YOURPATH\ukintrates.txt;
i3m = inter[.,1];
i10y = inter[.,2];
ysl = i10y - i3m;       /* slope of yield curve */



{ydep,yexp} = embed(dgdp,4,1,1);    /* ydep: Q2 56 to Q2 06 */
                                    /* yexp:    Q1 56 ... Q2 55
                                                ...
                                                Q1 06 ... Q2 05 */

{rub,xexp} = embed(ysl,12,1,1);    /* rub: not needed */
                                    /* xexp:    Dec 60 ... Jan 60 
                                                ...
                                                Sep 06 ... Oct 05 */

/* Setup the observations such that the first observation has the following data:

dep var         exp var
dgdp(Q1 61)     dgdp(Q4 60) ... dgdp(Q1 60) ysl(Dec 60) ... ysl(Jan 06) 
...
dgdp(Q2 06)     dgdp(Q1 06) ... dgdp(Q2 05) ysl(Mar 06) ... ysl(Apr 05) 
*/

ydep = trimr(ydep,19,0);
yexp = trimr(yexp,19,0);
xexp = trimr(xexp,0,6);

ncstar_maxlag = 4;

sel = seqa(1,3,rows(ydep)); /* pick end of quarter */
xexp = xexp[sel,.];

d73 = zeros(rows(ydep),1); 
d73[49] = 1;                    /* blip dummy for Q1 1973, placed relative to 1st obs Q1 61 */
data = ydep;
xall = yexp[.,1:3]~xexp[.,1:2]; /* max 4 lags for dgdp and 12 for ysl, restricted for NCSTR */
xx = xexp[.,1:ncstar_maxlag];               /* matrix from which trans var is calculated - for NCSTR restricted to 4 lags */
xdet = ones(rows(data),1)~d73;  /* deterministic variables, incl. const. */

xall_1saf = xall[2:rows(xall),.];               /* xall for forecasting */
xall_1saf = xdet[2:rows(xall),.]~xall_1saf;
xall      = xall[1:rows(xall)-1,.];             /* xall for estimation */
xx_1saf   = xx[2:rows(xx),.];                   /* xx for forecasting */
xx        = xx[1:rows(xx)-1,.];                 /* xx for estimation */
xdet_1saf = xdet[2:rows(xdet),.];                 /* xdet for forecasting */
xdet      = xdet[1:rows(xdet)-1,.];               /* xdet for estimation */
   
tt = rows(xall);        /* total number of estimation observations */
t0 = 156;               /* initial estimation sample */
t_oos = 1;              /* forecasting length 
                        should be = 1 if exogenous RHS variables */
nfore = (tt-t0)/t_oos; 
ns = 1;

ncstar_rss =  zeros(nfore,1);
ncstar_f = zeros(nfore,t_oos);
ncstar_ci5 = zeros(nfore,t_oos); ncstar_ci25 = ncstar_ci5; ncstar_ci75 = ncstar_ci5; ncstar_ci95 = ncstar_ci5;
ncstar_omega = zeros(nfore,2*ncstar_maxlag);

do while ns <= nfore;       /* start recursive loop */

  print "";
  print "****** forecast period = " ns;
  print "";

  y00 = data[1:t0+(ns-1)*t_oos];                    /* estimation sample */
  xall00 = xall[1:t0+(ns-1)*t_oos,.];               /* x for linear part, estimation sample */
  xx00 = xx[1:t0+(ns-1)*t_oos,.];                   /* xx is needed for the calculation of the transition variable */
  xall00_1saf = xall_1saf[t0+(ns-1)*t_oos,.];       /* extract the relevant matrices for 1 step ahead forecasts */
  xx00_1saf = xx_1saf[t0+(ns-1)*t_oos,.];
  xdet00 = xdet[1:t0+(ns-1)*t_oos,.];
  xdet00_1saf = xdet_1saf[t0+(ns-1)*t_oos,.];

  
  print "=====================================";
  print "== Linear Model - Robust se =========";
  print "=====================================";
  print "coefs    -   t-tests";
  test = hcse(y00,(xdet00~xall00));
  print test[.,1]~test[.,1]./test[.,2];


                /* Variable set-up */
  tvar = xx00;          /* transition variable */
  y0 = y00;
  z0 = xall00;  /* linear variables */
  sellin = seqa(1,1,5); /* used to draw linear variables */




  /*=============== SETUP FOR NCSTR ================*/

  struct svp {
    matrix om;
    matrix gam;
    matrix c; 
  };

  struct svp svtest;
  struct PV p0;               @ optimisation parameter structure @
  p0 = pvCreate;
  struct PV p1;               @ optimisation parameter structure @
  p1 = pvCreate;
  struct cmlmtControl con0;   @ control structure for optimisation @
  con0 = cmlmtControlCreate;
  struct cmlmtResults out1;   @ results structure for optimisation @
  struct cmlmtResults out2;   @ results structure for optimisation @
  #include c:\YOURPATH\ncstar_proc_ex.src;     /* all relevant procs are in here, needs the above structure definitions */


  /* Rather than NCSTAR Step1 we use the same linear model as for WSTAR, i.e the linear model estimated with yexp[.,1:3]~xexp[.,1:2] */

  /* Linearity test in NCSTAR */
  {selnl_1,pmin_1,pbon_1} = ncstar_step2_ex(y00,xall00[.,1:3],xx00);

  print "===================================================";
  print "H0: Linearity";
  print "Min Pvalue (Bon Pvalue) = " pmin_1 "(" pbon_1 ")";
  print "Selected nonlinear lags for tvar: " selnl_1;
  print "===================================================";


  /* Setup 1 neuron NCSTR model */

  tvar = xx00[.,selnl_1];                      /* nonlinear vars */

  struct DS d0;                           /* structure to hold data */
  d0 = dsCreate;
  d0.dataMatrix = y00;

  struct DS d1;                           /* structure to hold data */
  d1 = dsCreate;
  d1.dataMatrix = xdet00~xall00;

  struct DS d2;                           /* structure to hold data */
  d2 = dsCreate;
  d2.dataMatrix = tvar;

  d0 = d0|d1|d2;                          /* merge data structures, d0.1: y, d0.2: z, d0.3: x */

  /* 1 neuron NCSTAR */
  out1 = ncstar_n1(d0);

  if cols(d0[3].dataMatrix)>1;
    omega_hat = pvUnpack(out1.par,1);
    svtest.om = abs(sqrt(1-omega_hat'omega_hat)|omega_hat);    @ add first element if required, see (8) in MV(2005) @ 
  else;       
    svtest.om = 1;                @ 1 nonlin variable case @
  endif;
  svtest.gam = pvUnpack(out1.par,2);
  svtest.c   = pvUnpack(out1.par,3);




  {testsig, testres} = cmlncs1_full_report(svtest, d0, 0);
  print "===================================================";
  print "sighat (1 neuron NCSTR) = " testsig;

  {selnl_2,pmin_2,pbon_2} = ncstar_step3_ex(testres,xall00[.,1:3],xx00);    @ test for neglected nonlinearity @
  print "===================================================";
  print "H0: Linearity of NCSTAR 1 neuron resids";
  print "Min Pvalue (Bon Pvalue) = " pmin_2 "(" pbon_2 ")";
  print "Selected nonlinear lags for tvar: " selnl_2;
  print "===================================================";



  if pbon_2 < 0.05;

    struct DS d4;                           /* structure to hold data */
    d4 = dsCreate;
    d4.dataMatrix = xx00[.,selnl_2];        /* nonlinear vars for 2nd neuron */
    d0 = d0|d4;

    out2 = ncstar_n2(d0);
 
    k1 = cols(d0[3].dataMatrix);
    k2 = cols(d0[4].dataMatrix);
    svtest.gam = pvUnpack(out1.par,2);
    svtest.c   = pvUnpack(out1.par,3);
    omega = pvUnpack(out1.par,1);

    couom = 1;        @ counts used elements in omega @
    if k1 == 1;         @ only one element in x of d[3] -> omega = 1 @
      o1    = 1;
      couom = couom + k1;
      if k2 == 1;
        o2    = 1;
      else;
        o2    = omega[couom:(couom+k2-1-1)];
        couom = couom + k2 - 1 - 1;
        o2 = abs(sqrt(1-o2'o2))|o2;     @ see restriction (8) in MV(2005) @
      endif;
    else;
      o1 = omega[1:k1-1];
      couom = couom + (k1-1);
      if k2 == 1;
        o2    = omega[couom];
        couom = couom + k2 - 1;
      else;
        o2    = omega[couom:(couom+k2-1-1)];
        couom = couom + k2 - 1 - 1;
        o2 = abs(sqrt(1-o2'o2))|o2;     @ see restriction (8) in MV(2005) @
      endif;
      o1 = abs(sqrt(1-o1'o1))|o1;       @ see restriction (8) in MV(2005) @
    endif;
 
    svtest.om = o1|o2;
  
    {test,res2} = cmlncs2_full_report(svtest, d0, 0);
    print "===================================================";
    print "sighat (2 neuron NCSTR) = " test;

    {ey2,qy2} = NCSTARfore_ex(svtest, d0, sellin, selnl_1, selnl_2, 1);
    print "===================================================";
    print "Forecast (expected) = " ey2;
    print "Forecast (quantiles)= " qy2;
    print "===================================================";

    ncstar_omega[ns,selnl_1] = o1'; ncstarb_omega[ns,ncstar_maxlag+selnl_2] = o2';
    ncstar_rss[ns,.] = res2'res2;
    ncstar_f[ns,.] = ey2;
    ncstar_ci5[ns,.] = qy2[1,.]; ncstar_ci25[ns,.] = qy2[2,.]; ncstar_ci75[ns,.] = qy2[3,.]; ncstar_ci95[ns,.] = qy2[4,.];

  else;

    {ey1,qy1} = NCSTARfore_ex(svtest, d0, xall00_1saf, xx00_1saf, selnl_1, 0, 1);
    print "===================================================";
    print "Forecast (expected) = " ey1;
    print "Forecast (quantiles)= " qy1;
    print "===================================================";

    ncstar_omega[ns,selnl_1] = svtest.om';
    ncstar_rss[ns,.] = testres'testres;
    ncstar_f[ns,.] = ey1;
    ncstar_ci5[ns,.] = qy1[1,.]; ncstar_ci25[ns,.] = qy1[2,.]; ncstar_ci75[ns,.] = qy1[3,.]; ncstar_ci95[ns,.] = qy1[4,.];

  endif;


  ns = ns+1;           
endo;     /* end of forecasting loop */


/*===============
Function Section
================*/

/*==================================================
	hcse(y,x)	
	this procedure calculates heteroscedasticity
	consistent standard errors for regression
	coefficients
	input:	(i)		y, dep data
			(ii)	x, indep data
	output:	(i) 	kx4 matrix of
					- parameter estimates
					- Newey West se
					- White se
					- OLS se
	The lag length for the Newey-West window is
	set to int(4(n/100)^(2/9))
==================================================*/
proc (1) = hcse(y,x);

local n,k,b,e,ex,g,i,l,la,w,za,f,inf,v,nwerr,wherr,olserr;

n = rows(x);
k = cols(x);
b = y/x;
e = y - x*b;
ex = e.*x;
la = int(4*(n/100)^(2/9));
				/* Newey-West */
for i(0,la,1);
  l = i;
  w = (la+1-i)/(la+1);
  za = trimr(ex,l,0)'trimr(ex,0,l);
  if i==0; g = za;
  else; g = g+w*(za+za');
  endif;
endfor;

f = x'*x;
inf = inv(f);
v = inf*g*inf;
nwerr = (diag(v))^.5;

				/* White */

wherr = inf*(ex'ex)*inf;
wherr = (diag(wherr))^.5;

olserr=(diag(inv(x'x)*stdc(e)^2))^.5;
retp(b~nwerr~wherr~olserr);

endp;



/*=======================================================
	proc to embed data
	input: 	(i)		z, data vector
			(ii)	k, embed dimension
					if scalar it creates all lags up to k
					if nx1 vector it creates the n lags
			(iii)	tau, embed lag
			(iv)	seq, if 1 then lags = 1 .. k
						 if 0 then lags = {k}
	output:	(i)		y, image
			(ii)	x, embeded lags
    x only contains a constant if k == 0;
	Ralf Becker, 18.5.2000
=======================================================*/

proc (2) = embed(z,k,tau,seq);

local yx,j,x,y,max;


if (seq==1);
  yx = trimr(z,k*tau,0);
  for j(tau,k*tau,tau);
    yx = yx ~ trimr(z,(k*tau-j),j); 
  endfor;
  y = yx[.,1];
  if (k ne 0); x = yx[.,2:cols(yx)];
  else; x = ones(rows(y),1);
  endif;
else;		/* if selected lags */
  max = maxc(k);
  yx = trimr(z,max*tau,0);
  for j(1,rows(k),1);
    yx = yx ~ trimr(z,(max*tau-k[j]*tau),k[j]*tau); 
  endfor;
  y = yx[.,1];
endif;

if (k ne 0); x = yx[.,2:cols(yx)];
else; x = ones(rows(y),1);
endif;

retp(y,x);
endp;
